package projekt.enigma.model;

import java.util.*;
import java.util.Map.Entry;

/**
 * Diese Klasse definiert die Zugehoerigkeit zwischen den Buchstaben A-Z und dem Morsecode.
 * Die Funktionen sollen Texte von und zu Morsecode konvertieren.
 */
public class Morsecode {

	/**
	 * Anlegen der codeMap als TreeMap
	 */
	private TreeMap<String, String> codeMap;

	/**
	 * Im Konstruktor werden alle verfuegbaren Buchstaben (A-Z) und deren zugehoerigen
	 * Morsecodes, der codeMap hinzugefuegt.
	 */
	public Morsecode() {
		codeMap = new TreeMap<>();
		this.codeMap.put("A", ".-");
		this.codeMap.put("B", "-...");
		this.codeMap.put("C", "-.-.");
		this.codeMap.put("D", "-..");
		this.codeMap.put("E", ".");
		this.codeMap.put("F", "..-.");
		this.codeMap.put("G", "--.");
		this.codeMap.put("H", "....");
		this.codeMap.put("I", "..");
		this.codeMap.put("J", ".---");
		this.codeMap.put("K", "-.-");
		this.codeMap.put("L", ".-..");
		this.codeMap.put("M", "--");
		this.codeMap.put("N", "-.");
		this.codeMap.put("O", "---");
		this.codeMap.put("P", ".--.");
		this.codeMap.put("Q", "--.-");
		this.codeMap.put("R", ".-.");
		this.codeMap.put("S", "...");
		this.codeMap.put("T", "-");
		this.codeMap.put("U", "..-");
		this.codeMap.put("V", "...-");
		this.codeMap.put("W", ".--");
		this.codeMap.put("X", "-..-");
		this.codeMap.put("Y", "-.--");
		this.codeMap.put("Z", "--..");
		this.codeMap.put(" ", "-...-");
		this.codeMap.put("1", ".----");
		this.codeMap.put("2", "..---");
		this.codeMap.put("3", "...--");
		this.codeMap.put("4", "....-");
		this.codeMap.put("5", ".....");
		this.codeMap.put("6", "-....");
		this.codeMap.put("7", "--...");
		this.codeMap.put("8", "---..");
		this.codeMap.put("9", "----.");
		this.codeMap.put("0", "-----");
	}

	/**
	 * Diese Funktion ueberprueft ob der uebergebene Buchstabe in der codeMap enthalten ist.
	 * Wenn er enthalten ist, wird der zugehoerige Morsecode zurueck gegeben, ansonsten null.
	 *
	 * @param buchstabe : char : Zu uebersetzender Buchstabe
	 * @return String : Der zugehoerige Morsecode aus der codeMap
	 */
	private String getMorsecode(char buchstabe) {
		return this.codeMap.getOrDefault(String.valueOf(buchstabe).toUpperCase(), null);
	}

	/**
	 * Der uebergebene String wird in seine einzelnen Zeichen (char) zerlegt und an die Funktion getMorsecode
	 * uebergeben und der zugehoerige Morsecode raus gesucht.
	 * Mittels eines StringBuilders werden die Ergebnisse zu einem Rückgabe String zusammen gesetzt.
	 * Wenn kein zugehoeriger Code in der codeMap gefunden wurde, wird das Zeichen ignoriert da es bei der
	 * Enigma nicht vorhanden war.
	 *
	 * @param input : String : Zu konvertierender Text
	 * @return String : Der in Morsecode konvertierte Text
	 */
	public String convertBuchstabeToMorsecode(String input) {
		StringBuilder sb = new StringBuilder();
		String morsecode;

		// Durch alle char im uebergebenen String loopen
		for (int i = 0; i < input.length(); i++) {
			morsecode = this.getMorsecode(input.charAt(i));
			if (morsecode != null) {
				sb.append(morsecode).append(' ');
			}
		}

		return sb.toString();
	}

	/**
	 * Diese Funktion ueberprueft ob der uebergebene Morsecode in der codeMap enthalten ist.
	 * Wenn er enthalten ist, wird der zugehoerige Buchstabe zurueck gegeben, ansonsten null
	 *
	 * @param morsecode : String : Der zu uebersetzende Morsecode
	 * @return : String : Der zugehoerige Buchstabe aus der codeMap
	 */
	private String getBuchstabe(String morsecode) {
		String result = null;

		// Loopt duch die codeMap und erstellt daraus ein Set
		for (Entry<String, String> entry : this.codeMap.entrySet()) {
			if (entry.getValue().equals(morsecode)) {
				result = entry.getKey();
			}
		}

		return result;
	}

	/**
	 * Der uebergebene String wird in ein String Array zerlegt. Als Trennzeichen wird das Leerzeichen genutzt.
	 * Anschließend wird das Array durchlaufen und die Elemente an die Funktion getBuchstabe uebergeben um den
	 * zugehoerigen Buchstaben zu erhalten.
	 * Mittels eines StringBuilders werden die Ergebnisse zu einem Rueckgabe String zusammengesetzt.
	 * Wenn kein zugehoeriger Buchstabe in der codeMap gefunden wurde, wird das Zeichen ignoriert da es bei der
	 * Enigma nicht vorhanden war.
	 *
	 * @param input : String : Zu konvertierender Morsecode
	 * @return String : Der in Text konvertierte Morsecode
	 */
	public String convertMorsecodeToBuchstabe(String input) {
		StringBuilder sb = new StringBuilder();
		String buchstabe;

		String[] morsecodes = input.split(" ");

		// Durch alle char im uebergebenen String loopen
		for (String morsecode : morsecodes) {
			buchstabe = this.getBuchstabe(morsecode);
			if (buchstabe != null) {
				sb.append(buchstabe);
			}
		}

		return sb.toString();
	}
}
